package com.megabras.bluelogg.bluetooth;

import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattDescriptor;
import android.bluetooth.BluetoothProfile;
import android.content.Context;
import android.os.Handler;

import com.megabras.modbus.MUtils;
import com.megabras.modbus.Message;
import com.megabras.modbus.Modbus;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.Objects;
import java.util.Queue;
import java.util.UUID;

public class BluetoothBLE extends BluetoothDev
{

    private BluetoothGatt
            gatt;

    private final BluetoothGattCallback
            gattCallback;

    private final UUID
            uuidPassThroughService        = UUID.fromString( "0000ffe0-0000-1000-8000-00805f9b34fb" ),
            uuidSendCharacteristc         = UUID.fromString( "0000ffe9-0000-1000-8000-00805f9b34fb" ),
            uuidReadCharacteristc         = UUID.fromString( "0000ffe4-0000-1000-8000-00805f9b34fb" ),
            uuidReadDescriptor            = UUID.fromString( "00002902-0000-1000-8000-00805f9b34fb" ),
            uuidModuleParamService        = UUID.fromString( "0000ff90-0000-1000-8000-00805f9b34fb" ),
            uuidConnIntervalCharacteristc = UUID.fromString( "0000ff92-0000-1000-8000-00805f9b34fb" );

    private BluetoothGattCharacteristic
            writeCharacteristic,
            notifyCharacteristic,
            connIntervalCharacteristic;

    private final Queue<byte[]>
            fifo_tx_packages = new LinkedList<>();

    private int
            CONNECTION_STATE = BluetoothProfile.STATE_DISCONNECTED;

    public BluetoothBLE( Context context, BluetoothDevice device, Handler handler, Message messageStatus )
    {
        super( context, device, handler, messageStatus );

        gattCallback = new BluetoothGattCallback()
        {
            @Override
            public void onConnectionStateChange( BluetoothGatt bluetoothGatt, int status, int newState )
            {
                CONNECTION_STATE = newState;

                if ( newState == BluetoothProfile.STATE_CONNECTED )
                {
                    setState( STATE_CONNECTED );

                    if ( bluetoothGatt.getServices().isEmpty() )
                    {
                        handler.postDelayed( () ->
                        {
                            if ( CONNECTION_STATE == BluetoothProfile.STATE_CONNECTED )
                            {
                                bluetoothGatt.discoverServices();
                            }
                        }, 200 );
                    }
                }
                else if ( newState == BluetoothProfile.STATE_DISCONNECTED )
                {
                    stop();
                }
            }

            @Override
            public void onServicesDiscovered( BluetoothGatt bluetoothGatt, int status )
            {
                if ( status == BluetoothGatt.GATT_SUCCESS )
                {
                    if ( !gatt.getServices().isEmpty() )
                    {
                        writeCharacteristic        = gatt.getService( uuidPassThroughService ).getCharacteristic( uuidSendCharacteristc );
                        notifyCharacteristic       = gatt.getService( uuidPassThroughService ).getCharacteristic( uuidReadCharacteristc );
                        connIntervalCharacteristic = gatt.getService( uuidModuleParamService ).getCharacteristic( uuidConnIntervalCharacteristc );

                        if ( writeCharacteristic != null && notifyCharacteristic != null  )
                        {
                            gatt.setCharacteristicNotification( notifyCharacteristic, true );
                            BluetoothGattDescriptor descriptor = notifyCharacteristic.getDescriptor( uuidReadDescriptor );
                            descriptor.setValue( BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE );
                            gatt.writeDescriptor( descriptor );
                        }
                    }
                }
            }

            @Override
            public void onDescriptorWrite( BluetoothGatt bluetoothGatt, BluetoothGattDescriptor descriptor, int status )
            {
                if ( status == BluetoothGatt.GATT_SUCCESS )
                {
                    connIntervalCharacteristic.setValue( "0" );
                    gatt.writeCharacteristic( connIntervalCharacteristic );
                }
            }

            @Override
            public void onCharacteristicWrite( BluetoothGatt bluetoothGatt, BluetoothGattCharacteristic characteristic, int status )
            {
                if ( status == BluetoothGatt.GATT_SUCCESS )
                {
                    if ( characteristic.getUuid().equals( uuidConnIntervalCharacteristc ) )
                    {
                        mHandler.postDelayed( runFIFO, CYCLE_TIME );
                    }

                    if ( !fifo_tx_packages.isEmpty() )
                    {
                        writeCharacteristic.setValue( fifo_tx_packages.remove() );
                        gatt.writeCharacteristic( writeCharacteristic );
                        if ( fifo_tx_packages.isEmpty() )
                        {
                            int id = Objects.requireNonNull( FIFO_MESSAGE.peek() ).getID();
                            mHandler.obtainMessage
                            (
                                    MESSAGE_WRITE,
                                    -1,
                                    id,
                                    Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage().replace( ":", "" ).replace( "\r\n", "" )
                            ).sendToTarget();
                        }
                    }
                }
            }

            @Override
            public void onCharacteristicChanged( BluetoothGatt bluetoothGatt, BluetoothGattCharacteristic characteristic )
            {
                String readMessage = new String( characteristic.getValue(), 0, characteristic.getValue().length );
                rx += readMessage;
                current_time = 0;
            }
        };
    }

    @Override
    public synchronized void connect( boolean secure )
    {
        super.connect( secure );
        gatt = device.connectGatt( context, false, gattCallback );

        FIFO_MESSAGE = new LinkedList<>();
        runFIFO = new Runnable()
        {
            public void run()
            {
                clycleMessage();
                mHandler.postDelayed( this, CYCLE_TIME );
            }
        };

        mHandler.removeCallbacks( runFIFO );
    }

    public synchronized void stop()
    {
        mHandler.removeCallbacks( runFIFO );

        if ( gatt != null )
        {
            gatt.close();
            gatt = null;
        }

        setState( STATE_CLOSED );
    }

    private void clycleMessage()
    {
        if ( idle ) // COMUNICATION: IDLE
        {
            if ( !FIFO_MESSAGE.isEmpty() )
            {
                idle         = false;
                rx           = "";
                write( (Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage()).getBytes() );
                current_time = 0;
            }
        }
        else // COMUNICATION: WAITING FOR RESPONSE
        {
            if ( checkResponseWithFrames( rx ) || checkResponseWithoutFrames( rx ) )
            {
                rx = rx.replace( Modbus.START_FRAME,"").replace(Modbus.START_FRAME,"");

                if ( FIFO_MESSAGE.peek() == null )
                {
                    return;
                }

                String tx = Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage().replace(Modbus.START_FRAME,"").replace(Modbus.START_FRAME,"");

                boolean isExpectedAnswer = false;
                String rxFunctionCode    = rx.substring( 2, 4 );
                String txFunctionCode    = tx.substring( 2, 4 );

                if ( rxFunctionCode.equals( txFunctionCode ) )
                {
                    if ( rxFunctionCode.equals( "03" ) )
                    {
                        int txSize = MUtils.hexToInt( tx.substring( 8, 12 ) );
                        int rxSize = MUtils.hexToInt( rx.substring( 4, 6 ) ) / 2;
                        if ( txSize == rxSize )
                            isExpectedAnswer = true;
                    }
                    else
                    {
                        String txAddress = tx.substring( 4, 8 );
                        String rxAddress = rx.substring( 4, 8 );
                        if ( rxAddress.equals( txAddress ) )
                            isExpectedAnswer = true;
                    }
                }

                if ( isExpectedAnswer && FIFO_MESSAGE.peek() != null )
                {
                    mHandler.obtainMessage( MESSAGE_READ, 0, Objects.requireNonNull( FIFO_MESSAGE.peek() ).getID(), rx ).sendToTarget();
                    FIFO_MESSAGE.remove();
                    current_time = 0;
                    idle = true;
                    rx = "";
                }
            }
        }

        // COMUNICATION: Rx TIMEOUT and CHECK CONNECTION
        if ( current_time == TIME_OUT || current_time == (TIME_OUT * 2) || current_time == (TIME_OUT * 4) ) // Retry last command OR ask status
        {
            if ( FIFO_MESSAGE.isEmpty() )
            {
                FIFO_MESSAGE.add( messageStatus );
            }
            rx = "";
            idle = false;
            if ( FIFO_MESSAGE.peek() != null )
            {
                write( (Objects.requireNonNull( FIFO_MESSAGE.peek() ).getMessage()).getBytes() );
            }
        }
        else if ( current_time == (TIME_OUT * 6) )
        {
            setState( STATE_NOT_RESPONDING );
            idle = true;
            rx = "";
            current_time = 0;
        }
        current_time += CYCLE_TIME;
    }

    public void write( byte[] buffer )
    {
        try
        {
            if ( gatt != null && CONNECTION_STATE == BluetoothProfile.STATE_CONNECTED && FIFO_MESSAGE.peek() != null )
            {
                int id = Objects.requireNonNull( FIFO_MESSAGE.peek() ).getID();

                if ( buffer.length <= 20 )
                {
                    writeCharacteristic.setValue( buffer );
                    gatt.writeCharacteristic( writeCharacteristic );

                    mHandler.obtainMessage( MESSAGE_WRITE, -1, id, new String( buffer ).replace( ":", "" ).replace( "\r\n", "" ) ).sendToTarget();
                }
                else
                {
                    fifo_tx_packages.clear();

                    int packSize = 20;
                    for ( int i = 0; i < buffer.length; i += packSize  )
                    {
                        int to = ( i + packSize ) < buffer.length ? i + packSize : buffer.length;
                        fifo_tx_packages.add( Arrays.copyOfRange( buffer, i, to ) );
                    }

                    writeCharacteristic.setValue( fifo_tx_packages.remove() );
                    gatt.writeCharacteristic( writeCharacteristic );
                }
            }
        }
        catch ( Exception ignore)
        {
        }
    }
}
